//
// Time-stamp: <30.10.2003 14:33:37h liekweg>
//
// This library is distributed under the BSD license.
// See license.txt for more information.
//
// $Id$
//

package dot;

/**
   <P>Root class of all dot node attributes.</P>

   @since   Sun Oct 26 16:52:40 2003
   @author  Florian Liekweg <TT>&lt;liekweg@ipd.info.uni-karlsruhe.de&gt;</TT>, Universität Karlsruhe (TH), Germany
 */

public interface DotNodeAttr extends DotAttr {}

