/*************************************************************************
 * Copyright (c) 2011 AT&T Intellectual Property 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/org/documents/epl-2.0/EPL-2.0.html
 *
 * Contributors: Details at https://graphviz.org
 *************************************************************************/


#include "config.h"

#include <gvc/gvplugin_layout.h>

typedef enum { LAYOUT_DOT, } layout_type;

extern void dot_layout(graph_t * g);
extern void dot_cleanup(graph_t * g);

static gvlayout_engine_t dotgen_engine = {
    dot_layout,
    dot_cleanup,
};

static gvlayout_features_t dotgen_features = {
    LAYOUT_USES_RANKDIR,
};

gvplugin_installed_t gvlayout_dot_layout[] = {
    {LAYOUT_DOT, "dot", 0, &dotgen_engine, &dotgen_features},
    {0, NULL, 0, NULL, NULL}
};
